/**
 * Service
 * @class ApEntityService
 * @param {Object} store - Redux store
 * @param {Object} [options={}] - Optional settings
 */
'use strict'

const { ApService } = require('apeman-service-base')
const abind = require('abind')
const uuid = require('uuid')
const co = require('co')

let toFieldName = (...names) => names.filter(Boolean).join(':')

let asList = (entityName) => toFieldName(entityName, 'list')
let asOne = (entityName) => toFieldName(entityName, 'one')
let asResource = (entityName) => toFieldName(entityName, 'resource')

let isEmpty = (value) => (typeof value === 'undefined') || (value === null) || (value === '') || (value === false)

/** @lends ApEntityService */
class ApEntityService extends ApService {
  constructor (store, options = {}) {
    super(store)
    const s = this
    abind(s, { proto: ApEntityService.prototype })
    abind(s)
    s._listCache = {}
  }

  // ---------------
  // Getter Methods
  // ---------------

  /**
   * Get entity as list
   * @param {string} entityName - Name of entity.
   * @returns {Object} - Entry values
   */
  getEntityList (entityName) {
    const s = this
    let { ids, counts, cacheKey } = s.get(asList(entityName)) || {}
    if (isEmpty(ids)) {
      return null
    }
    if (!counts) {
      return null
    }

    // Restore cache
    {
      let { _listCache } = s
      let cache = _listCache[ entityName ]
      if (cache) {
        let valid = cache.cacheKey === cacheKey
        if (valid) {
          return cache // Return save instance for shallow compare
        }
      }
    }

    let resources = s.getEntityResources(entityName)
    let { total } = counts
    let result = {
      counts,
      entities: ids.map((id) => {
        let found = resources.hasOwnProperty(id)
        if (!found) {
          console.warn(`Entity resource not found with id: ${id}`)
        }
        return resources[ id ]
      }),
      more: ids.length < total
    }
    // Save to cache
    {
      let { _listCache } = s
      _listCache[ entityName ] = result
      s._listCache = _listCache
    }
    return result
  }

  /**
   * Get detailed entity
   * @param {string} entityName - Name of entity.
   * @returns {*}
   */
  getEntityOne (entityName) {
    const s = this
    let resources = s.getEntityResources(entityName)
    let id = s.get(asOne(entityName))
    if (isEmpty(id)) {
      return null
    }
    return resources[ id ]
  }

  /**
   * Get entity resources.
   * @param {string} entityName - Name of entity.
   * @returns {Object}
   */
  getEntityResources (entityName) {
    const s = this
    return s.get(asResource(entityName)) || {}
  }

  // ---------------
  // Action Methods
  // ---------------

  /**
   * Update entity resources.
   * @param {string} entityName - Name of entity.
   * @param {Array.<Object>|Object} entities - Entities to push
   * @returns {Promise}
   */
  updateEntityResources (entityName, entities) {
    const s = this
    s._listCache = {}// Remove cache
    return co(function * () {
      let resources = [].concat(entities).reduce((resources, entity) => Object.assign(resources, {
        [entity.id]: entity
      }), {})
      s.store.update(asResource(entityName), resources)
    })
  }

  /**
   * Set entity list.
   * @param {string} entityName - Name of entity.
   * @param {Array.<Object>} entities - Entities to push
   * @param {Object} counts - Count info
   * @returns {Promise}
   */
  setEntityList (entityName, entities, counts) {
    const s = this
    return co(function * () {
      yield s.updateEntityResources(entityName, entities)
      s.store.set(asList(entityName), {
        ids: entities.map((entity) => entity.id),
        counts,
        cacheKey: uuid.v4()
      })
    })
  }

  /**
   * Push to entity list.
   * @param {string} entityName - Name of entity.
   * @param {Array.<Object>} entities - Entities to push
   * @param {Object} counts - Count info
   * @returns {Promise}
   */
  pushEntityList (entityName, entities, counts) {
    const s = this
    return co(function * () {
      let {
        entities: currentEntities = [],
        counts: currentCounts = {}
      } = s.getEntityList(entityName) || {}
      yield s.setEntityList(
        entityName,
        [ ...currentEntities, ...entities ],
        Object.assign({}, currentCounts, counts)
      )
    })
  }

  /**
   * Unshift to entity list.
   * @param {string} entityName - Name of entity.
   * @param {Array.<Object>} entities - Entities to unshift
   * @param {Object} counts - Count info
   * @returns {Promise}
   */
  unshiftEntityList (entityName, entities, counts) {
    const s = this
    return co(function * () {
      let {
        entities: currentEntities = [],
        counts: currentCounts = {}
      } = s.getEntityList(entityName) || {}
      yield s.setEntityList(
        entityName,
        [ ...entities, ...currentEntities ],
        Object.assign({}, currentCounts, counts)
      )
    })
  }

  /**
   * Set entity one
   * @param {string} entityName - Name of entity.
   * @param {Object} entity - Entity to set.
   * @returns {Promise}
   */
  setEntityOne (entityName, entity) {
    const s = this
    return co(function * () {
      let { id } = entity
      yield s.updateEntityResources(entityName, entity)
      s.store.set(asOne(entityName), id)
    })
  }

  /**
   * Clear entity list
   * @param {string} entityName - Name of entity.
   * @returns {Promise}
   */
  clearEntityList (entityName) {
    const s = this
    return co(function * () {
      s.store.del(asList(entityName))
    })
  }

  /**
   * Clear entity one
   * @param {string} entityName - Name of entity.
   * @returns {Promise}
   */
  clearEntityOne (entityName) {
    const s = this
    return co(function * () {
      s.store.del(asOne(entityName))
    })
  }

  /**
   * Clear entity resource
   * @param {string} entityName - Name of entity.
   * @returns {Promise}
   */
  clearEntityResource (entityName) {
    const s = this
    return co(function * () {
      s.store.del(asResource(entityName))
    })
  }

  /**
   * Clear entity one
   * @param {string} entityName - Name of entity.
   * @returns {Promise}
   */
  purgeEntity (entityName) {
    const s = this
    return co(function * () {
      yield s.clearEntityOne(entityName)
      yield s.clearEntityList(entityName)
      yield s.clearEntityResource(entityName)
    })
  }

  /**
   * Invalidate entity cache
   * @param {string} entityName - Name of entity.
   * @returns {Promise}
   */
  invalidateEntityCache (entityName) {
    const s = this
    let { _listCache } = s
    delete _listCache[ entityName ]
  }
}

module.exports = ApEntityService
